@@ -52,11 +52,10 @@ mail_item_free (MailItem *item)
 gboolean
 fetch_props_to_mail_item_cb (FetchItemsCallbackData *item_data, gpointer data)
 {
-	long *flags;
-	struct FILETIME *delivery_date;
-	const gchar *msg_class;
-	NTTIME ntdate;
-	ExchangeMAPIStream *body;
+	long *flags = NULL;
+	struct FILETIME *delivery_date = NULL;
+	const gchar *msg_class = NULL;
+	ExchangeMAPIStream *body = NULL;
 
 	MailItem *item;
 	MailItem **i = (MailItem **)data;
@@ -101,7 +100,7 @@ fetch_props_to_mail_item_cb (FetchItemsCallbackData *item_data, gpointer data)
 	}
 
 	item->is_cal = FALSE;
-	if (g_str_has_prefix (msg_class, IPM_SCHEDULE_MEETING_PREFIX)) {
+	if (msg_class && g_str_has_prefix (msg_class, IPM_SCHEDULE_MEETING_PREFIX)) {
 		guint8 *appointment_body_str = (guint8 *) exchange_mapi_cal_util_camel_helper (item_data->conn, item_data->fid, item_data->mid, NULL, msg_class,
 												item_data->streams, item_data->recipients, item_data->attachments);
 
@@ -132,15 +131,17 @@ fetch_props_to_mail_item_cb (FetchItemsCallbackData *item_data, gpointer data)
 	}
 
 	if (delivery_date) {
+		NTTIME ntdate = { 0 };
+
 		ntdate = delivery_date->dwHighDateTime;
 		ntdate = ntdate << 32;
 		ntdate |= delivery_date->dwLowDateTime;
 		item->header.recieved_time = nt_time_to_unix(ntdate);
 	}
 
-	if ((*flags & MSGFLAG_READ) != 0)
+	if (flags && (*flags & MSGFLAG_READ) != 0)
 		item->header.flags |= CAMEL_MESSAGE_SEEN;
-	if ((*flags & MSGFLAG_HASATTACH) != 0)
+	if (flags && (*flags & MSGFLAG_HASATTACH) != 0)
 		item->header.flags |= CAMEL_MESSAGE_ATTACHMENTS;
 
 	item->attachments = item_data->attachments;
@@ -260,9 +261,6 @@ mapi_mail_get_item_prop_list (ExchangeMapiConnection *conn, mapi_id_t fid, TALLO
 		PR_BODY,
 		PR_BODY_UNICODE,
 		PR_HTML,
-		/*Fixme : If this property is fetched, it garbles everything else. */
-		/*PR_BODY_HTML, */
-		/*PR_BODY_HTML_UNICODE, */
 
 		PR_DISPLAY_TO_UNICODE,
 		PR_DISPLAY_CC_UNICODE,
@@ -296,7 +294,6 @@ mapi_mail_get_item_prop_list (ExchangeMapiConnection *conn, mapi_id_t fid, TALLO
 	return exchange_mapi_utils_add_props_to_props_array (mem_ctx, props, item_props, G_N_ELEMENTS (item_props));
 }
 
-
 static void
 mapi_mime_set_recipient_list (CamelMimeMessage *msg, MailItem *item)
 {
@@ -366,7 +363,6 @@ static void
 mapi_mime_set_msg_headers (ExchangeMapiConnection *conn, CamelMimeMessage *msg, MailItem *item)
 {
 	gchar *temp_str = NULL;
-	const gchar *from_email;
 	time_t recieved_time;
 	CamelInternetAddress *addr = NULL;
 	gint offset = 0;
@@ -380,11 +376,11 @@ mapi_mime_set_msg_headers (ExchangeMapiConnection *conn, CamelMimeMessage *msg,
 
 		stream = camel_stream_mem_new_with_buffer (item->header.transport_headers, strlen (item->header.transport_headers));
 		parser = camel_mime_parser_new ();
-		camel_mime_parser_init_with_stream (parser, stream);
+		camel_mime_parser_init_with_stream (parser, stream, NULL);
 		camel_mime_parser_scan_from (parser, FALSE);
 		g_object_unref (stream);
 
-		if (camel_mime_part_construct_from_parser (part, parser) != -1) {
+		if (camel_mime_part_construct_from_parser (part, parser, NULL) != -1) {
 			struct _camel_header_raw *h;
 
 			for (h = part->headers; h; h = h->next) {
@@ -418,9 +414,11 @@ mapi_mime_set_msg_headers (ExchangeMapiConnection *conn, CamelMimeMessage *msg,
 
 	if (item->header.from) {
 		if ((item->header.from_type != NULL) && !g_utf8_collate (item->header.from_type, "EX")) {
-			from_email = exchange_mapi_connection_ex_to_smtp (conn, item->header.from_email);
+			gchar *from_email;
+
+			from_email = exchange_mapi_connection_ex_to_smtp (conn, item->header.from_email, NULL);
 			g_free (item->header.from_email);
-			item->header.from_email = g_strdup (from_email);
+			item->header.from_email = from_email;
 		}
 
 		item->header.from_email = item->header.from_email ?
@@ -569,7 +567,7 @@ mapi_mime_build_multipart_alternative (MailItem *item, GSList *body_parts, GSLis
 	while (body_parts) {
 		const ExchangeMAPIStream *stream = (ExchangeMAPIStream *) body_parts->data;
 		part = camel_mime_part_new ();
-		if ((stream->proptag == PR_HTML || stream->proptag == PR_BODY_HTML_UNICODE)
+		if ((stream->proptag == PR_HTML)
 		    && inline_attachs) {
 			CamelMultipart *m_related;
 			m_related = mapi_mime_build_multipart_related (item, stream,
@@ -582,6 +580,8 @@ mapi_mime_build_multipart_alternative (MailItem *item, GSList *body_parts, GSLis
 
 		camel_multipart_add_part (m_alternative, part);
 		g_object_unref (part);
+
+		body_parts = body_parts->next;
 	}
 
 	return m_alternative;
@@ -605,10 +605,57 @@ mapi_mime_build_multipart_mixed (CamelMultipart *content, GSList *attachs)
 	return m_mixed;
 }
 
+static gboolean
+is_apple_attachment (ExchangeMAPIAttachment *attach, guint32 *data_len, guint32 *resource_len)
+{
+	gboolean is_apple = FALSE;
+	ExchangeMAPIStream *enc_stream = exchange_mapi_util_find_stream (attach->streams, PR_ATTACH_ENCODING);
+	guint8 apple_enc_magic[] = { 0x2A, 0x86, 0x48, 0x86, 0xF7, 0x14, 0x03, 0x0B, 0x01 };
+
+	if (enc_stream && enc_stream->value && enc_stream->value->len == G_N_ELEMENTS (apple_enc_magic)) {
+		gint idx;
+
+		is_apple = TRUE;
+		for (idx = 0; idx < enc_stream->value->len && is_apple; idx++) {
+			is_apple = apple_enc_magic[idx] == enc_stream->value->data[idx];
+		}
+	} else {
+		const struct Binary_r *bin = exchange_mapi_util_find_SPropVal_array_propval (attach->lpProps, PR_ATTACH_ENCODING);
+		if (bin && bin->cb == G_N_ELEMENTS (apple_enc_magic)) {
+			gint idx;
+
+			is_apple = TRUE;
+			for (idx = 0; idx < bin->cb && is_apple; idx++) {
+				is_apple = apple_enc_magic[idx] == bin->lpb[idx];
+			}
+		}
+	}
+
+	if (is_apple) {
+		/* check boundaries too */
+		ExchangeMAPIStream *data_stream = exchange_mapi_util_find_stream (attach->streams, PR_ATTACH_DATA_BIN);
+
+		is_apple = data_stream && data_stream->value && data_stream->value->len > 128;
+
+		if (is_apple) {
+			const guint8 *bin = data_stream->value->data;
+
+			/* in big-endian format */
+			*data_len = (bin[83] << 24) | (bin[84] << 16) | (bin[85] << 8) | (bin[86]);
+			*resource_len = (bin[87] << 24) | (bin[88] << 16) | (bin[89] << 8) | (bin[90]);
+
+			/* +/- mod 128 (but the first 128 is a header length) */
+			is_apple = 128 + *data_len + *resource_len <= data_stream->value->len && bin[1] < 64;
+		}
+	}
+
+	return is_apple;
+}
+
 /*Takes raw attachment streams and converts to MIME Parts. Parts are added to
   either inline / non-inline lists.*/
 static void
-mapi_mime_classify_attachments (GSList *attachments, GSList **inline_attachs, GSList **noninline)
+mapi_mime_classify_attachments (ExchangeMapiConnection *conn, mapi_id_t fid, GSList *attachments, GSList **inline_attachs, GSList **noninline)
 {
 	for (;attachments != NULL; attachments = attachments->next) {
 		ExchangeMAPIAttachment *attach = (ExchangeMAPIAttachment *)attachments->data;
@@ -617,6 +664,8 @@ mapi_mime_classify_attachments (GSList *attachments, GSList **inline_attachs, GS
 		CamelContentType *content_type;
 		CamelMimePart *part;
 		const uint32_t *ui32;
+		gboolean is_apple;
+		guint32 apple_data_len = 0, apple_resource_len = 0;
 
 		stream = exchange_mapi_util_find_stream (attach->streams, PR_ATTACH_DATA_BIN);
 
@@ -624,16 +673,7 @@ mapi_mime_classify_attachments (GSList *attachments, GSList **inline_attachs, GS
 			continue;
 		}
 
-		part = camel_mime_part_new ();
-
-		filename = (const gchar *) exchange_mapi_util_find_SPropVal_array_propval(attach->lpProps,
-											 PR_ATTACH_LONG_FILENAME_UNICODE);
-
-		if (!(filename && *filename))
-			filename = (const gchar *) exchange_mapi_util_find_SPropVal_array_propval(attach->lpProps,
-												 PR_ATTACH_FILENAME_UNICODE);
-		camel_mime_part_set_filename(part, g_strdup(filename));
-		camel_content_type_set_param (((CamelDataWrapper *) part)->mime_type, "name", filename);
+		is_apple = is_apple_attachment (attach, &apple_data_len, &apple_resource_len);
 
 		/*Content-Type*/
 		ui32 = (const uint32_t *) exchange_mapi_util_find_SPropVal_array_propval (attach->lpProps, PR_ATTACH_METHOD);
@@ -645,19 +685,107 @@ mapi_mime_classify_attachments (GSList *attachments, GSList **inline_attachs, GS
 				mime_type = "application/octet-stream";
 		}
 
-		camel_mime_part_set_content (part, (const gchar *) stream->value->data, stream->value->len, mime_type);
+		if (is_apple) {
+			mime_type = "application/applefile";
+		} else if (strstr (mime_type, "apple") != NULL) {
+			mime_type = "application/octet-stream";
+		}
+
+		part = camel_mime_part_new ();
+
+		filename = (const gchar *) exchange_mapi_util_find_SPropVal_array_propval(attach->lpProps,
+											 PR_ATTACH_LONG_FILENAME_UNICODE);
+
+		if (!(filename && *filename))
+			filename = (const gchar *) exchange_mapi_util_find_SPropVal_array_propval(attach->lpProps,
+												 PR_ATTACH_FILENAME_UNICODE);
+		camel_mime_part_set_filename (part, filename);
+		camel_content_type_set_param (((CamelDataWrapper *) part)->mime_type, "name", filename);
+
+		if (is_apple) {
+			ExchangeMAPIStream *strm;
+			CamelMultipart *mp;
+			uint32_t proptag;
+			gchar *apple_filename;
+
+			mp = camel_multipart_new ();
+			camel_data_wrapper_set_mime_type (CAMEL_DATA_WRAPPER (mp), "multipart/appledouble");
+			camel_multipart_set_boundary (mp, NULL);
 
-		content_type = camel_mime_part_get_content_type (part);
-		if (content_type && camel_content_type_is (content_type, "text", "*"))
-			camel_mime_part_set_encoding (part, CAMEL_TRANSFER_ENCODING_QUOTEDPRINTABLE);
-		else if (!ui32 || *ui32 != ATTACH_EMBEDDED_MSG)
 			camel_mime_part_set_encoding (part, CAMEL_TRANSFER_ENCODING_BASE64);
 
+			strm = NULL;
+			proptag = exchange_mapi_connection_resolve_named_prop (conn, fid, PidNameAttachmentMacInfo, NULL);
+			if (proptag != MAPI_E_RESERVED)
+				strm = exchange_mapi_util_find_stream (attach->streams, proptag);
+			if (!strm)
+				strm = exchange_mapi_util_find_stream (attach->streams, PidNameAttachmentMacInfo);
+
+			if (strm && strm->value && strm->value->len > 0) {
+				camel_mime_part_set_content (part, (const gchar *) strm->value->data, strm->value->len, mime_type);
+			} else {
+				/* RFC 1740 */
+				guint8 header[] = {
+					0x00, 0x05, 0x16, 0x07, /* magic */
+					0x00, 0x02, 0x00, 0x00, /* version */
+					0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* filler */
+					0x00, 0x01, /* number of entries */
+					0x00, 0x00, 0x00, 0x02, /* entry ID - resource fork */
+					0x00, 0x00, 0x00, 0x26, /* entry offset - 38th byte*/
+					0x00, 0x00, 0x00, 0x00  /* entry length */
+				};
+
+				GByteArray *arr = g_byte_array_sized_new (apple_resource_len + G_N_ELEMENTS (header));
+
+				header[34] = (apple_resource_len >> 24) & 0xFF;
+				header[35] = (apple_resource_len >> 16) & 0xFF;
+				header[36] = (apple_resource_len >>  8) & 0xFF;
+				header[37] = (apple_resource_len      ) & 0xFF;
+
+				g_byte_array_append (arr, header, G_N_ELEMENTS (header));
+				g_byte_array_append (arr, stream->value->data + 128 + apple_data_len + (apple_data_len % 128), apple_resource_len);
+
+				camel_mime_part_set_content (part, (const gchar *) arr->data, arr->len, mime_type);
+
+				g_byte_array_free (arr, TRUE);
+			}
+
+			camel_multipart_add_part (mp, part);
+			g_object_unref (part);
+
+			part = camel_mime_part_new ();
+
+			apple_filename = g_strndup ((gchar *)stream->value->data + 2, stream->value->data[1]);
+			camel_mime_part_set_filename (part, (apple_filename && *apple_filename) ? apple_filename : filename);
+			g_free (apple_filename);
+
+			mime_type = exchange_mapi_util_find_SPropVal_array_namedid (attach->lpProps, conn, fid, PidNameAttachmentMacContentType);
+			if (!mime_type)
+				mime_type = "application/octet-stream";
+
+			camel_mime_part_set_content (part, (const gchar *) stream->value->data + 128, apple_data_len, mime_type);
+			camel_mime_part_set_encoding (part, CAMEL_TRANSFER_ENCODING_BASE64);
+			camel_multipart_add_part (mp, part);
+			g_object_unref (part);
+
+			part = camel_mime_part_new ();
+			camel_medium_set_content (CAMEL_MEDIUM (part), CAMEL_DATA_WRAPPER (mp));
+			g_object_unref (mp);
+		} else {
+			camel_mime_part_set_content (part, (const gchar *) stream->value->data, stream->value->len, mime_type);
+
+			content_type = camel_mime_part_get_content_type (part);
+			if (content_type && camel_content_type_is (content_type, "text", "*"))
+				camel_mime_part_set_encoding (part, CAMEL_TRANSFER_ENCODING_QUOTEDPRINTABLE);
+			else if (!ui32 || *ui32 != ATTACH_EMBEDDED_MSG)
+				camel_mime_part_set_encoding (part, CAMEL_TRANSFER_ENCODING_BASE64);
+		}
+
 		/*Content-ID*/
 		content_id = (const gchar *) exchange_mapi_util_find_SPropVal_array_propval(attach->lpProps,
 											   PR_ATTACH_CONTENT_ID);
 		/* TODO : Add disposition */
-		if (content_id) {
+		if (content_id && !is_apple) {
 			camel_mime_part_set_content_id (part, content_id);
 			*inline_attachs = g_slist_append (*inline_attachs, part);
 		} else
@@ -686,9 +814,9 @@ mapi_mail_item_to_mime_message (ExchangeMapiConnection *conn, MailItem *item)
 
 	mapi_mime_set_recipient_list (msg, item);
 	mapi_mime_set_msg_headers (conn, msg, item);
-	mapi_mime_classify_attachments (attach_list, &inline_attachs, &noninline_attachs);
+	mapi_mime_classify_attachments (conn, item->fid, attach_list, &inline_attachs, &noninline_attachs);
 
-	build_alternative = (g_slist_length (item->msg.body_parts) > 1) && inline_attachs;
+	build_alternative = g_slist_length (item->msg.body_parts) > 1;
 	build_related = !build_alternative && inline_attachs;
 
 	if (build_alternative) {
